from collections import OrderedDict

import mongoengine as mongo
from mongoengine.base import BaseField, ComplexBaseField

from .misc import Methods, url_join

base_types_mapping = {
    "ObjectIdField": {"type": "integer", "format": "int64"},
    "BooleanField": {"type": "boolean"},
    "StringField": {"type": "string"},
    "IntField": {"type": "number"},
    "ListField": {"type": "array", "items": {}}
}


def __field_mapping__(self):
    # Define mappings between Mongoengine fields and OpenAPI 3

    try:
        # base_types_mapping is static and without the dict the references will
        # confuse the generator thus dict is required to create a new item
        toRet = dict(base_types_mapping[self.__class__.__name__])

        if self.__class__.__name__ in ["ListField"]:  # Complex field? Generate mapping recursively
            toRet["items"] = self.field._gen_schema()

        return toRet
    except:
        # Return StringField if you don't find the exact mapping
        return base_types_mapping["StringField"]


def generic_endpoint_spec(class_name, operation, require_id):
    operation2verb = {
        "create": "created",
        "read": "read",
        "update": "updated",
        "delete": "deleted",
    }

    toRet = OrderedDict()

    toRet["tags"] = [class_name]
    toRet["summary"] = f"{operation} a {class_name}"

    if require_id:
        toRet["operationId"] = f"{class_name}_{operation}_id"

        toRet["parameters"] = [{}]
        toRet["parameters"][0]["name"] = "id"
        toRet["parameters"][0]["in"] = "path"
        toRet["parameters"][0]["description"] = f"ID/pk of {class_name} to return"
        toRet["parameters"][0]["required"] = "true"
        toRet["parameters"][0]["schema"] = {}
        toRet["parameters"][0]["schema"]["type"] = "string"
        # toRet["parameters"][0]["schema"]["format"] = "int64"

    else:
        toRet["operationId"] = f"{class_name}_{operation}"

    if operation not in ["read", "delete"]:
        toRet["requestBody"] = {}
        toRet["requestBody"]["description"] = f"{class_name} object that needs to be {operation2verb[operation]}"
        toRet["requestBody"]["required"] = True
        toRet["requestBody"]["content"] = {}
        toRet["requestBody"]["content"]["application/json"] = {"schema": {"$ref": f"#/components/schemas/{class_name}"}}
        # toRet["requestBody"]["content"]["application/xml"] = {"schema": {"$ref": "#/components/schemas/{}".format(class_name) }}

    toRet["responses"] = {}
    toRet["responses"]["200"] = {
        "description": "Successful operation",
        "content": {
            "application/json": {
                "schema": {"$ref": f"#/components/schemas/{class_name}"}
            }
        },
    }
    toRet["responses"]["405"] = {"description": "Invalid input"}

    return toRet


def gen_openapi_path_spec(class_name, methods, path):
    toRet = OrderedDict()

    path_no_id = url_join([path, class_name])
    path_with_id = url_join([path, class_name, "{id}"])

    toRet[path_with_id] = OrderedDict()
    toRet[path_no_id] = OrderedDict()

    if Methods.post in methods:
        toRet[path_no_id]["post"] = generic_endpoint_spec(class_name, "create", require_id=False)

    if Methods.get in methods:
        toRet[path_no_id]["get"] = generic_endpoint_spec(class_name, "read", require_id=False)
        toRet[path_with_id]["get"] = generic_endpoint_spec(class_name, "read", require_id=True)

    if Methods.update in methods:
        toRet[path_with_id]["put"] = generic_endpoint_spec(class_name, "update", require_id=True)

    if Methods.delete in methods:
        toRet[path_with_id]["delete"] = generic_endpoint_spec(class_name, "delete", require_id=True)

    if len(toRet[path_with_id]) == 0:
        del toRet[path_with_id]

    if len(toRet[path_no_id]) == 0:
        del toRet[path_no_id]

    return toRet


@classmethod
def __gen_root_document_OpenAPI_spec(_class):
    properties = OrderedDict()

    for key in _class._fields_ordered:  # <mongoengine.base.fields.ObjectIdField object at 0x10a2e8208>
        # print("---------------------------------")
        # print(key) #name
        # print(self._fields[key])
        # print(type(self._fields[key]).__name__) # StringFiled
        # print(self._fields[key].__class__.__name__) # StringFiled
        # print(vars(type(self._fields[key])) ) # StringFiled
        # if key=="id": continue
        # try:
        field_name = _class._fields[key].name
        properties[field_name] = _class._fields[key]._gen_schema()

    return OrderedDict({"type": "object", "properties": properties})


def __start_openapi3_dict__(url, title, version, description, **kwargs):

    json_result = OrderedDict()
    json_result["openapi"] = "3.0.3"
    json_result["info"] = {
        "title": title,
        "version": version,
        "description": description,
    }
    json_result["servers"] = [{"url": url}]
    json_result["paths"] = OrderedDict()
    json_result["components"] = {"schemas": OrderedDict()}

    return json_result


def mokeypatch_mongoengine():
    # Monkey Patch schema to generate class definition

    BaseField._gen_schema = __field_mapping__

    mongo.Document._gen_schema = __gen_root_document_OpenAPI_spec

    # mongo.ObjectIdField._gen_schema = __field_mapping__
    # mongo.StringField._gen_schema = __field_mapping__
    # # mongo.URLField
    # # mongo.EmailField
    # # mongo.IntField._gen_schema = _gen_schema_int
    # # mongo.LongField
    # # mongo.FloatField
    # # mongo.DecimalField._gen_schema = _gen_schema_dict
    # mongo.BooleanField._gen_schema = __field_mapping__
    # # mongo.DateTimeField._gen_schema = _gen_schema_date_time
    # # mongo.ComplexDateTimeField._gen_schema = _gen_schema_cdate
    # # mongo.EmbeddedDocumentField
    # # mongo.GenericEmbeddedDocumentField
    # # mongo.DynamicField
    # mongo.ListField._gen_schema = __field_mapping__
    # # mongo.SortedListField
    # # mongo.DictField._gen_schema = _gen_schema_dict
    # # mongo.MapField
    # # mongo.ReferenceField
    # # mongo.GenericReferenceField
    # # mongo.BinaryField
    # # mongo.FileField
    # # mongo.ImageField
    # # mongo.SequenceField
    # # mongo.GeoPointField
    # # mongo.PointField
    # # mongo.LineStringField
    # # mongo.PolygonField
    # # mongo.GridFSError
    # # mongo.GridFSProxy
    # # mongo.ImageGridFsProxy
    # # mongo.ImproperlyConfigured ?????


# Stolen from https://github.com/tiangolo/fastapi/blob/5614b94ccc9f72f1de2f63aae63f5fe90b86c8b5/fastapi/openapi/docs.py


from typing import Any, Dict, Optional, OrderedDict


# Generate basic html map with all possible routes inside flask
def get_map_html(flask_api) -> str:

    endpoint_routes = {}
    for rule in flask_api.url_map.iter_rules():
        methods = ','.join(sorted(rule.methods))
        endpoint_routes[f"{rule.rule}{methods}"] = rule.rule, methods

    html = """
    <!DOCTYPE html>
    <html>
    <head>

    </head>
    <body>
    <div>
    <ul>
    """

    for t in sorted(endpoint_routes.keys()):
        route = endpoint_routes[t]
        print(route)
        tmp = route[0].replace("<", "&lt;").replace(">", "&gt;")
        html += f"<li><a href='{tmp}'>{tmp} {route[1]}</a></li>"

    html += """
    </ul>
    </div>
    </body>
    </html>
    """

    return html


def get_swagger_ui_html(
    *,
    openapi_url: str,
    title: str,
    swagger_js_url: str = "https://cdn.jsdelivr.net/npm/swagger-ui-dist@3/swagger-ui-bundle.js",
    swagger_css_url: str = "https://cdn.jsdelivr.net/npm/swagger-ui-dist@3/swagger-ui.css",
    swagger_favicon_url: str = "https://fastapi.tiangolo.com/img/favicon.png",
    oauth2_redirect_url: Optional[str] = None,
    init_oauth: Optional[Dict[str, Any]] = None,
    **kwargs
) -> str:  # HTMLResponse:

    html = f"""
    <!DOCTYPE html>
    <html>
    <head>
    <link type="text/css" rel="stylesheet" href="{swagger_css_url}">
    <link rel="shortcut icon" href="{swagger_favicon_url}">
    <title>{title}</title>
    </head>
    <body>
    <div id="swagger-ui">
    </div>
    <script src="{swagger_js_url}"></script>
    <!-- `SwaggerUIBundle` is now available on the page -->
    <script>
    const ui = SwaggerUIBundle({{
        url: '{openapi_url}',
    """

    if oauth2_redirect_url:
        html += f"oauth2RedirectUrl: window.location.origin + '{oauth2_redirect_url}',"

    html += """
        dom_id: '#swagger-ui',
        presets: [
        SwaggerUIBundle.presets.apis,
        SwaggerUIBundle.SwaggerUIStandalonePreset
        ],
        layout: "BaseLayout",
        deepLinking: true,
        showExtensions: true,
        showCommonExtensions: true
    })"""

    # if init_oauth:
    #     html += f"""
    #     ui.initOAuth({json.dumps(jsonable_encoder(init_oauth))})
    #     """

    html += """
    </script>
    </body>
    </html>
    """
    # return HTMLResponse(html)
    return html


def get_redoc_html(
    *,
    openapi_url: str,
    title: str,
    redoc_js_url: str = "https://cdn.jsdelivr.net/npm/redoc@next/bundles/redoc.standalone.js",
    redoc_favicon_url: str = "https://fastapi.tiangolo.com/img/favicon.png",
    **kwargs
    # with_google_fonts: bool = True,
) -> str:  # HTMLResponse:
    html = f"""
    <!DOCTYPE html>
    <html>
    <head>
    <title>{title}</title>
    <!-- needed for adaptive design -->
    <meta charset="utf-8"/>
    <meta name="viewport" content="width=device-width, initial-scale=1">
    """
    # if with_google_fonts:
    #     html += """
    # <link href="https://fonts.googleapis.com/css?family=Montserrat:300,400,700|Roboto:300,400,700" rel="stylesheet">
    # """
    html += f"""
    <link rel="shortcut icon" href="{redoc_favicon_url}">
    <!--
    ReDoc doesn't change outer page styles
    -->
    <style>
      body {{
        margin: 0;
        padding: 0;
      }}
    </style>
    </head>
    <body>
    <redoc spec-url="{openapi_url}"></redoc>
    <script src="{redoc_js_url}"> </script>
    </body>
    </html>
    """
    # return HTMLResponse(html)
    return html

#
# def get_swagger_ui_oauth2_redirect_html() -> HTMLResponse:
#     html = """
#     <!DOCTYPE html>
#     <html lang="en-US">
#     <body onload="run()">
#     </body>
#     </html>
#     <script>
#         'use strict';
#         function run () {
#             var oauth2 = window.opener.swaggerUIRedirectOauth2;
#             var sentState = oauth2.state;
#             var redirectUrl = oauth2.redirectUrl;
#             var isValid, qp, arr;
#             if (/code|token|error/.test(window.location.hash)) {
#                 qp = window.location.hash.substring(1);
#             } else {
#                 qp = location.search.substring(1);
#             }
#             arr = qp.split("&")
#             arr.forEach(function (v,i,_arr) { _arr[i] = '"' + v.replace('=', '":"') + '"';})
#             qp = qp ? JSON.parse('{' + arr.join() + '}',
#                     function (key, value) {
#                         return key === "" ? value : decodeURIComponent(value)
#                     }
#             ) : {}
#             isValid = qp.state === sentState
#             if ((
#             oauth2.auth.schema.get("flow") === "accessCode"||
#             oauth2.auth.schema.get("flow") === "authorizationCode"
#             ) && !oauth2.auth.code) {
#                 if (!isValid) {
#                     oauth2.errCb({
#                         authId: oauth2.auth.name,
#                         source: "auth",
#                         level: "warning",
#                         message: "Authorization may be unsafe, passed state was changed in server Passed state wasn't returned from auth server"
#                     });
#                 }
#                 if (qp.code) {
#                     delete oauth2.state;
#                     oauth2.auth.code = qp.code;
#                     oauth2.callback({auth: oauth2.auth, redirectUrl: redirectUrl});
#                 } else {
#                     let oauthErrorMsg
#                     if (qp.error) {
#                         oauthErrorMsg = "["+qp.error+"]: " +
#                             (qp.error_description ? qp.error_description+ ". " : "no accessCode received from the server. ") +
#                             (qp.error_uri ? "More info: "+qp.error_uri : "");
#                     }
#                     oauth2.errCb({
#                         authId: oauth2.auth.name,
#                         source: "auth",
#                         level: "error",
#                         message: oauthErrorMsg || "[Authorization failed]: no accessCode received from the server"
#                     });
#                 }
#             } else {
#                 oauth2.callback({auth: oauth2.auth, token: qp, isValid: isValid, redirectUrl: redirectUrl});
#             }
#             window.close();
#         }
#     </script>
#         """
#     return HTMLResponse(content=html)



from flask import Flask

from . import flask_endpoint, openapi_spec
from .misc import Methods, get_default_api_info, url_join


class FlaskRestGlue:
    def __init__(self, url="http://127.0.0.1:5000", path="/", spec_path=None, flask_app=None, api_info=None, gen_map=True):
        self._path = url_join([path])
        self._spec_path = spec_path
        self._flask_api = flask_app or Flask(__name__)
        self._pending_introspection = {}
        self._api_info = api_info
        self._gen_map = gen_map
        self._openapi_spec = None

        if self._spec_path:
            self._spec_path = url_join([spec_path])
        else:
            self._spec_path = url_join([self._path, "api.json"])

        if not self._api_info:
            self._api_info = get_default_api_info(base_url=url, path_spec=self._spec_path)

    def rest_model(self, name=None, methods=Methods.all, *args, **kwargs):
        def inner(_class):
            """
            Store items pending introspection
            """
            class_name_lower = (name or f"{_class.__name__}").lower()

            self._pending_introspection[_class] = methods, class_name_lower

            return _class

        return inner

    def build(self):
        # Generate api_doc and endpoint
        openapi_spec.mokeypatch_mongoengine()

        openapi_spec_dict = openapi_spec.__start_openapi3_dict__(**self._api_info)

        for cls in self._pending_introspection.keys():

            methods, class_name_lower = self._pending_introspection[cls]

            # Generate spec
            schema = cls._gen_schema()

            spec_path = openapi_spec.gen_openapi_path_spec(class_name_lower, methods, self._path)

            openapi_spec_dict["components"]["schemas"][class_name_lower] = schema
            openapi_spec_dict["paths"].update(spec_path)

            # Generate endpoint for each class
            flask_endpoint.build_class_endpoint(self._flask_api, cls, class_name_lower, methods, self._path)

        flask_endpoint.build_flask_openapi_spec_endpoint(self._flask_api, openapi_spec_dict, self._spec_path, self._api_info, self._path, self._gen_map)

        self._openapi_spec = openapi_spec_dict

        self._flask_api.config["JSON_SORT_KEYS"] = False

        return self._flask_api

    def run(self, *args, **kwargs):

        flask_api = self.build()

        flask_api.run(*args, **kwargs)



class Methods:

    get = "GET"
    put = "PUT"
    post = "POST"
    delete = "DELETE"
    update = "UPDATE"

    all = [get, put, post, delete, update]


def get_default_api_info(base_url, path_spec):
    default_api_info = {
        "title": "Flask_Rest_Glue",
        "version": "0.0.1",
        "description": "N/A",
        "url": f"{base_url}",
        "openapi_url": f"{base_url}{path_spec}"
    }
    return default_api_info


def url_join(pieces):

    # pieces = ["", "//home", "/User/Desktop", "file.txt"]
    out = '/'.join(s.strip('/') for s in pieces)
    # '/home/User/Desktop/file.txt'

    if len(out) == 0:
        return "/"

    if out[0] != "/":
        out = f"/{out}"

    return out



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['flask_rest_glue']

package_data = \
{'': ['*']}

install_requires = \
['Flask>=1.1.2,<2.0.0',
 'mongoengine>=0.22.1,<0.23.0',
 'typer[all]>=0.3.2,<0.4.0']

extras_require = \
{':python_version < "3.8"': ['importlib_metadata>=1.6.0,<2.0.0']}

setup_kwargs = {
    'name': 'flask-restglue',
    'version': '0.0.3',
    'description': 'Flask-RestGlue integrates Flask, MongoDB, OpenAPI in a simple and elegant way',
    'long_description': '\n\n<!-- PROJECT HEADER/LOGO -->\n<!-- <br /> -->\n\n<p align="center">\n  <!--\n  <a href="https://github.com/abassel/Flask-RestGlue">\n    <img src="images/logo.png" alt="Logo" width="80" height="80">\n  </a>\n  -->\n  <h1 align="center">Flask-RestGlue(ALPHA)</h1>\n\n  <p align="center">\n    Integrates <b>Flask + MongoDB + OpenAPI</b> in a simple and elegant way!\n    <br />\n    <a href="https://abassel.github.io/Flask-RestGlue/"><strong>Explore the docs »</strong></a>\n    <br />\n    <br />\n    <a href="https://github.com/abassel/Flask-RestGlue#example">View Demo</a>\n    ·\n    <a href="https://github.com/abassel/Flask-RestGlue/issues">Report Bug</a>\n    ·\n    <a href="https://github.com/abassel/Flask-RestGlue/issues">Request Feature</a>\n  </p>\n\n  <p align="center">\n      <a href="https://github.com/abassel/Flask-RestGlue/actions?query=workflow%3Abuild"><img src="https://github.com/abassel/Flask-RestGlue/workflows/build/badge.svg?branch=master&event=push" alt="Stars Badge"/></a>\n      <a href="https://github.com/abassel/Flask-RestGlue/pulls"><img src="https://img.shields.io/github/issues-pr/abassel/Flask-RestGlue" alt="Pull Requests Badge"/></a>\n      <a href="https://github.com/abassel/Flask-RestGlue/issues"><img src="https://img.shields.io/github/issues/abassel/Flask-RestGlue" alt="Issues Badge"/></a>\n  </p>\n\n  <p align="center">\n   <a href="https://pypi.org/project/Flask-RestGlue/"><img src="https://img.shields.io/pypi/pyversions/Flask-RestGlue.svg" alt="Python Version"/></a>\n   <a href="https://github.com/abassel/Flask-RestGlue/releases"><img src="https://img.shields.io/pypi/v/Flask-RestGlue?color=green&label=version" alt="Version"/></a>\n   <a href="https://github.com/abassel/Flask-RestGlue/pulls?utf8=%E2%9C%93&q=is%3Apr%20author%3Aapp%2Fdependabot"><img src="https://img.shields.io/badge/dependencies-up%20to%20date-brightgreen.svg" alt="Dependencies Status"/></a>\n   <a href="https://github.com/PyCQA/bandit"><img src="https://img.shields.io/badge/security-bandit-green.svg" alt="Security: bandit"/></a>\n   <a href="https://github.com/abassel/Flask-RestGlue/blob/master/.pre-commit-config.yaml"><img src="https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white" alt="Pre-commit"/></a>\n   <a href="https://github.com/abassel/Flask-RestGlue/blob/master/LICENSE"><img src="https://img.shields.io/github/license/abassel/Flask-RestGlue" alt="License"/></a>\n  </p></p>\n\n\n<!-- TABLE OF CONTENTS -->\n## Contents\n<details>\n  <summary>Table of Contents</summary>\n  <ol>\n    <!--\n    <li>\n      <a href="#about-the-project">About The Project</a>\n      <ul>\n        <li><a href="#built-with">Built With</a></li>\n      </ul>\n    </li>\n    <li>\n      <a href="#getting-started">Getting Started</a>\n      <ul>\n        <li><a href="#prerequisites">Prerequisites</a></li>\n        <li><a href="#installation">Installation</a></li>\n      </ul>\n    </li>\n    -->\n    <li><a href="#example">Example</a></li>\n    <li><a href="#quick-start">Quick Start</a></li>\n    <li><a href="#references-notebook">References</a></li>\n    <!--<li><a href="#roadmap">Roadmap</a></li>\n    <li><a href="#contributing">Contributing</a></li>\n    <li><a href="#license">License</a></li>\n    <li><a href="#contact">Contact</a></li>\n    <li><a href="#acknowledgements">Acknowledgements</a></li>-->\n  </ol>\n</details>\n\n\n\n# Example\n\nFor a fullstack boilerplate, visit [https://github.com/abassel/Flask_RestGlue_Svelte_Docker](https://github.com/abassel/Flask_RestGlue_Svelte_Docker)\n\n```python\nimport mongoengine as mongo\nfrom flask_rest_glue import FlaskRestGlue\n\nmongo.connect("pyglue", host=\'localhost:27017\')\n\napi = FlaskRestGlue()\n\n\n@api.rest_model()\nclass User(mongo.Document):\n  # id = mongo.StringField(primary_key=True)\n  email = mongo.StringField(primary_key=True)\n  password = mongo.StringField()\n\n\napi.run()\n\n```\n\nGo to [http://127.0.0.1:5000/spec_doc](http://127.0.0.1:5000/spec_doc) or [http://127.0.0.1:5000/spec_rdoc](http://127.0.0.1:5000/spec_rdoc) to see the documentation bellow:\n\n![Swagger UI](https://abassel.github.io/Flask-RestGlue/swagger.png)\n\n![ReText UI](https://abassel.github.io/Flask-RestGlue/rdoc.png)\n\n### Expected output:\n\n```bash\ncurl -v -d \'{"email":"a@b.com","password":"xyz"}\' \\\n     -H "Content-Type: application/json" http://localhost:5000/user\n\n#< HTTP/1.0 200 OK\n#< Content-Type: application/json\n#< Content-Length: 45\n#<\n#{\n#  "_id": "a@b.com",\n#  "password": "xyz"\n#}\n\n\ncurl -v http://localhost:5000/users\n#< HTTP/1.0 200 OK\n#< Content-Type: application/json\n#< Content-Length: 57\n#<\n#[\n#  {\n#    "_id": "a@b.com",\n#    "password": "xyz"\n#  }\n#]\n\n\ncurl -v http://localhost:5000/user/a@b.com\n#< HTTP/1.0 200 OK\n#< Content-Type: application/json\n#< Content-Length: 45\n#<\n#{\n#  "_id": "a@b.com",\n#  "password": "xyz"\n#}\n\n\ncurl -v -X PUT -d \'{"password":"new_pass"}\' \\\n     -H "Content-Type: application/json" http://localhost:5000/user/a@b.com\n#< HTTP/1.0 200 OK\n#< Content-Type: application/json\n#< Content-Length: 50\n#<\n#{\n#  "_id": "a@b.com",\n#  "password": "new_pass"\n#}\n\n\ncurl -v -X DELETE http://localhost:5000/user/a@b.com\n#< HTTP/1.0 200 OK\n#< Content-Type: application/json\n#< Content-Length: 45\n#<\n#{\n#  "_id": "a@b.com",\n#  "password": "xyz"\n#}\n```\n\n\n# Quick Start\n\nRequires **docker** and **python 3.9**\n\n### 1 - install local MongoDB\n```bash\nmkdir -p ~/mongodata\n\ndocker run -d --rm -p 27017:27017 -v ~/mongodata:/data/db --name mongodb mongo\n```\n\n### 2 - Install this library\n```bash\npip install Flask-RestGlue\n```\n\n### 3 - Pull the code\n```bash\ncurl -s -O -L  https://github.com/abassel/Flask-RestGlue/blob/master/example/tut01_hello_world.py\ncurl -s -O -L  https://github.com/abassel/Flask-RestGlue/blob/master/example/tut01_hello_world.sh\n```\n\n### 4 - Run the code\n```bash\npython tut01_hello_world.py\n```\nIn another terminal window\n\n```bash\nbash tut01_hello_world.sh\n```\n\n## References :notebook:\n- [Flask_RestGlue_Svelte_Docker boilerplate](https://github.com/abassel/Flask_RestGlue_Svelte_Docker)\n- [OpenAPI 3.0.3](https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.3.md)\n- [Flask](http://flask.pocoo.org)\n- [Mongoengine](https://github.com/MongoEngine/mongoengine)\n',
    'author': 'Alexandre Bassel',
    'author_email': 'abassel@abassel.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/abassel/Flask-RestGlue',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import json

from flask import jsonify, request

from .misc import Methods, url_join
from .openapi_spec import get_map_html, get_redoc_html, get_swagger_ui_html


def build_flask_openapi_spec_endpoint(flask_api, spec, path_spec, api_info, root_path, gen_map):

    @flask_api.route(path_spec)
    def json_spec():
        return jsonify(spec)

    @flask_api.route(path_spec.split(".json")[0] + "_doc")
    def swagger_html():
        return get_swagger_ui_html(**api_info)

    @flask_api.route(path_spec.split(".json")[0] + "_rdoc")
    def redoc_html():
        return get_redoc_html(**api_info)

    if gen_map:
        tmp_html = get_map_html(flask_api)

        @flask_api.route(root_path)
        def path_spec_html():
            return tmp_html


def build_class_endpoint(flask_api, _class, class_name, methods, path):

    # https://dev.to/paurakhsharma/flask-rest-api-part-1-using-mongodb-with-flask-3g7d

    root_path = url_join([path, class_name])
    root_path_pk = url_join([root_path, "<pk>"])

    if Methods.post in methods:
        @flask_api.route(root_path, endpoint=f"{class_name}_create", methods=["POST"])
        def obj_create_post():

            if hasattr(_class, "before_any"):
                _class.before_any(request)

            if hasattr(_class, "before_post"):
                _class.before_post(request)

            body = request.get_json()
            new_item = _class(**body).save()

            if hasattr(_class, 'custom_json'):
                return jsonify(new_item.custom_json())

            return jsonify(json.loads(new_item.to_json()))

    if Methods.get in methods:
        @flask_api.route(root_path, endpoint=f"{class_name}_read", methods=["GET"])
        def obj_read():

            if hasattr(_class, "before_any"):
                _class.before_any(request)

            if hasattr(_class, "before_get"):
                _class.before_get(request)

            items = _class.objects()

            if hasattr(_class, 'custom_json'):
                return jsonify([item.custom_json() for item in items])

            return jsonify(json.loads(items.to_json()))

    if Methods.get in methods:
        @flask_api.route(root_path_pk, endpoint=f"{class_name}_read_id", methods=["GET"])
        def obj_read(pk):

            if hasattr(_class, "before_any"):
                _class.before_any(request)

            if hasattr(_class, "before_get"):
                _class.before_get(request)

            item = _class.objects.get(pk=pk)

            if hasattr(_class, 'custom_json'):
                return jsonify(item.custom_json())

            return jsonify(json.loads(item.to_json()))

    if Methods.put in methods:
        @flask_api.route(root_path_pk, endpoint=f"{class_name}_update", methods=["PUT"])
        def obj_update(pk):

            if hasattr(_class, "before_any"):
                _class.before_any(request)

            if hasattr(_class, "before_put"):
                _class.before_put(request)

            body = request.get_json()
            item = _class.objects.get(pk=pk)
            item.update(**body)
            item.reload()

            if hasattr(_class, 'custom_json'):
                return jsonify(item.custom_json())

            return jsonify(json.loads(item.to_json()))

    if Methods.delete in methods:
        @flask_api.route(root_path_pk, endpoint=f"{class_name}_delete", methods=["DELETE"])
        def obj_delete(pk):

            if hasattr(_class, "before_any"):
                _class.before_any(request)

            if hasattr(_class, "before_delete"):
                _class.before_delete(request)

            item = _class.objects.get(pk=pk)
            item.delete()

            if hasattr(_class, 'custom_json'):
                return jsonify(item.custom_json())

            return jsonify(json.loads(item.to_json()))
